import time
from typing import List, Optional

from langchain_openai_api_bridge.core.types.openai import (
    OpenAIChatCompletionChoice,
    OpenAIChatCompletionObject,
    OpenAIChatCompletionUsage,
)


class ChatCompletionObjectFactory:
    def create(
        id: str,
        model: str,
        choices: List[OpenAIChatCompletionChoice] = [],
        usage: Optional[
            OpenAIChatCompletionUsage
        ] = OpenAIChatCompletionUsage.default(),
        object: str = "chat.completion",
        system_fingerprint: str = "",
        created: int = None,
    ) -> OpenAIChatCompletionObject:
        return OpenAIChatCompletionObject(
            id=id,
            object=object,
            created=created if created is not None else int(time.time()),
            model=model,
            system_fingerprint=system_fingerprint,
            choices=choices,
            usage=usage,
        )



from typing import Iterable, Literal, Optional, Union
from openai.types.beta.threads import MessageContentPartParam, message_create_params
from pydantic import BaseModel


class CreateThreadMessageDto(BaseModel):
    content: Union[str, Iterable[MessageContentPartParam]]
    role: Literal["user", "assistant"]
    attachments: Optional[Iterable[message_create_params.Attachment]] = None
    metadata: Optional[object] = None



from typing import List, Optional
from openai.types.beta import thread_create_params
from pydantic import BaseModel


class CreateThreadDto(BaseModel):
    messages: List[thread_create_params.Message] = []
    metadata: Optional[object] = None



from typing import AsyncIterator

from pydantic import BaseModel


async def ato_dict(async_iter: AsyncIterator[BaseModel]) -> AsyncIterator[dict]:
    async for obj in async_iter:
        yield obj.dict()



import json
from typing import AsyncIterator
from starlette.responses import StreamingResponse


class HttpStreamResponseAdapter:
    async def to_str_stream(
        self, chunk_stream: AsyncIterator[dict]
    ) -> AsyncIterator[str]:
        async for chunk in chunk_stream:
            yield self.__serialize_chunk(chunk)

        yield "data: [DONE]\n\n"

    def to_streaming_response(
        self, chunk_stream: AsyncIterator[dict]
    ) -> StreamingResponse:
        str_stream = self.to_str_stream(chunk_stream)
        return StreamingResponse(
            str_stream,
            headers={
                "Content-Type": "text/event-stream",
                "Transfer-Encoding": "chunked",
            },
            media_type="application/x-ndjson",
        )

    def __serialize_chunk(self, chunk):
        return f"data: {json.dumps(chunk)}\n\n"



from typing import Optional
from pydantic import BaseModel


class ThreadRunsDto(BaseModel):
    assistant_id: str
    thread_id: str = ""
    model: Optional[str] = None
    temperature: float = 0.7



from typing import Literal
from openai.types.beta.threads import Message, MessageDeleted


from langchain_openai_api_bridge.assistant.create_thread_message_api_dto import (
    CreateThreadMessageDto,
)
from langchain_openai_api_bridge.assistant.repository.message_repository import (
    MessageRepository,
)
from openai.pagination import SyncCursorPage


class AssistantMessageService:

    def __init__(
        self,
        message_repository: MessageRepository,
    ) -> None:
        self.message_repository = message_repository

    def create(self, thread_id: str, dto: CreateThreadMessageDto) -> Message:
        # Reference:
        # client.beta.threads.messages.create(
        #   "thread_abc123",
        #   role="user",
        #   content="How does AI work? Explain it in simple terms.",
        # )
        return self.message_repository.create(
            thread_id=thread_id, role=dto.role, content=dto.content
        )

    def list(
        self,
        thread_id: str,
        after: str = None,
        before: str = None,
        limit: int = None,
        order: Literal["asc", "desc"] = None,
    ) -> SyncCursorPage[Message]:
        # Reference:
        # client.beta.threads.messages.list("thread_abc123")
        return self.message_repository.listByPage(
            thread_id=thread_id, after=after, before=before, limit=limit, order=order
        )

    def retreive(self, message_id: str, thread_id: str) -> Message:
        # Reference:
        # client.beta.threads.messages.retrieve(
        #   message_id="msg_abc123",
        #   thread_id="thread_abc123",
        # )
        return self.message_repository.retreive(
            message_id=message_id, thread_id=thread_id
        )

    def delete(self, message_id: str, thread_id: str) -> MessageDeleted:
        # Reference:
        # client.beta.threads.messages.delete(
        #   message_id="msg_abc123",
        #   thread_id="thread_abc123",
        # )
        return self.message_repository.delete(
            message_id=message_id, thread_id=thread_id
        )



from typing import AsyncIterator, List, Optional
from langgraph.graph.graph import CompiledGraph
from langchain_openai_api_bridge.chat_completion.langchain_invoke_adapter import (
    LangchainInvokeAdapter,
)
from langchain_openai_api_bridge.chat_completion.langchain_stream_adapter import (
    LangchainStreamAdapter,
)
from langchain_openai_api_bridge.core.types.openai import OpenAIChatMessage
from langchain_openai_api_bridge.core.utils.pydantic_async_iterator import ato_dict


class ChatCompletionCompatibleAPI:

    @staticmethod
    def from_agent(
        agent: CompiledGraph, llm_model: str, system_fingerprint: Optional[str] = ""
    ):
        return ChatCompletionCompatibleAPI(
            LangchainStreamAdapter(llm_model, system_fingerprint),
            LangchainInvokeAdapter(llm_model, system_fingerprint),
            agent,
        )

    def __init__(
        self,
        stream_adapter: LangchainStreamAdapter,
        invoke_adapter: LangchainInvokeAdapter,
        agent: CompiledGraph,
    ) -> None:
        self.stream_adapter = stream_adapter
        self.invoke_adapter = invoke_adapter
        self.agent = agent

    def astream(self, messages: List[OpenAIChatMessage]) -> AsyncIterator[dict]:
        input = self.__to_input(messages)
        astream_event = self.agent.astream_events(
            input=input,
            version="v2",
        )
        return ato_dict(
            self.stream_adapter.ato_chat_completion_chunk_stream(astream_event)
        )

    def invoke(self, messages: List[OpenAIChatMessage]) -> dict:
        result = self.agent.invoke(
            input=self.__to_input(messages),
        )

        return self.invoke_adapter.to_chat_completion_object(result).dict()

    def __to_input(self, messages: List[OpenAIChatMessage]):
        return {
            "messages": [message.dict() for message in messages],
        }



from typing import AsyncIterator
from openai.types.beta import AssistantStreamEvent
from langchain_openai_api_bridge.assistant.adapter.langgraph_event_to_openai_assistant_event_stream import (
    LanggraphEventToOpenAIAssistantEventStream,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)
from langgraph.graph.graph import CompiledGraph

from langchain_openai_api_bridge.assistant.adapter.thread_to_langchain_input_messages_service import (
    ThreadToLangchainInputMessagesService,
)


class AssistantRunService:

    def __init__(
        self,
        thread_message_service: ThreadToLangchainInputMessagesService,
        stream_adapter: LanggraphEventToOpenAIAssistantEventStream,
    ) -> None:
        self.thread_message_service = thread_message_service
        self.stream_adapter = stream_adapter

    def astream(
        self, agent: CompiledGraph, dto: ThreadRunsDto
    ) -> AsyncIterator[AssistantStreamEvent]:

        input = self.thread_message_service.retreive_input_dict(thread_id=dto.thread_id)
        astream_events = agent.astream_events(
            input=input,
            version="v2",
        )

        return self.stream_adapter.to_openai_assistant_event_stream(
            astream_events=astream_events, dto=dto
        )



def to_string_content(content) -> str:
    if isinstance(content, str):
        return content

    elif isinstance(content, list):
        if len(content) == 0:
            return ""
        else:
            return content[0]["text"]



from .in_memory_message_repository import InMemoryMessageRepository
from .in_memory_run_repository import InMemoryRunRepository
from .in_memory_thread_repository import InMemoryThreadRepository

from .message_repository import MessageRepository
from .run_repository import RunRepository
from .thread_repository import ThreadRepository

__all__ = [
    "InMemoryMessageRepository",
    "InMemoryRunRepository",
    "InMemoryThreadRepository",
    "MessageRepository",
    "RunRepository",
    "ThreadRepository",
]



from langchain_core.runnables.schema import StreamEvent
from openai.types.beta import AssistantStreamEvent
from langchain_openai_api_bridge.assistant.adapter.openai_event_factory import (
    create_thread_message_completed,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)
from langchain_openai_api_bridge.assistant.repository.message_repository import (
    MessageRepository,
)

from openai.types.beta.threads import (
    TextContentBlock,
    Text,
)


class OnChatModelEndHandler:
    def __init__(
        self,
        thread_message_repository: MessageRepository,
    ):
        self.thread_message_repository = thread_message_repository

    def handle(
        self, event: StreamEvent, dto: ThreadRunsDto
    ) -> list[AssistantStreamEvent]:
        events = []
        run_id = event["run_id"]
        thread_id = dto.thread_id
        final_content_str = event["data"]["output"].content

        message = self.thread_message_repository.retreive_unique_by_run_id(
            run_id=run_id, thread_id=thread_id
        )

        if message is None:
            return events

        message.content = [
            TextContentBlock(
                type="text", text=Text(value=final_content_str, annotations=[])
            )
        ]

        message.status = "completed"
        completed_message = self.thread_message_repository.update(message)

        events.append(create_thread_message_completed(message=completed_message))

        return events



from typing import List
from langchain_openai_api_bridge.assistant.repository.message_repository import (
    MessageRepository,
)
from langchain_openai_api_bridge.core.types.openai.message import OpenAIChatMessage


class ThreadToLangchainInputMessagesService:
    def __init__(self, message_repository: MessageRepository):
        self.message_repository = message_repository

    def retreive_input(self, thread_id: str) -> List[OpenAIChatMessage]:
        thread_messages = self.message_repository.list(thread_id=thread_id)
        messages = [
            OpenAIChatMessage(
                role=message.role,
                content=message.content[0].text.value,
            )
            for message in thread_messages
        ]

        return messages

    def retreive_input_dict(self, thread_id: str) -> dict:
        messages = self.retreive_input(thread_id=thread_id)
        return {
            "messages": [message.dict() for message in messages],
        }



from typing import Callable, Optional, Union
from langchain_openai_api_bridge.core.agent_factory import AgentFactory
from langchain_openai_api_bridge.core.utils import TinyDIContainer


class LangchainOpenaiApiBridge:
    def __init__(
        self,
        agent_factory_provider: Union[Callable[[], AgentFactory], AgentFactory],
        tiny_di_container: Optional[TinyDIContainer] = None,
    ) -> None:
        self.tiny_di_container = tiny_di_container or TinyDIContainer()

        self.tiny_di_container.register(AgentFactory, to=agent_factory_provider)



from openai.types.beta.threads import Run
from abc import ABC, abstractmethod
from typing import List, Optional
from openai.types.beta.threads.run import RequiredAction, RunStatus, AssistantTool


class RunRepository(ABC):

    def __init__(
        self,
    ) -> None:
        pass

    @abstractmethod
    def create(
        self,
        assistant_id: str,
        thread_id: str,
        model: str,
        status: RunStatus,
        instructions: str = "",
        required_action: Optional[RequiredAction] = None,
        tools: List[AssistantTool] = [],
        parallel_tool_calls: bool = True,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
    ) -> Run:
        pass

    @abstractmethod
    def update(self, run: Run) -> Run:
        pass

    @abstractmethod
    def retreive(self, run_id: str) -> Run:
        pass

    @abstractmethod
    def delete(self, run: Optional[Run], run_id: Optional[str]) -> None:
        pass



from typing import Optional
from fastapi import HTTPException


def get_bearer_token(authorization: Optional[str] = None) -> str:
    if authorization and authorization.startswith("Bearer "):
        return authorization[len("Bearer "):]
    raise HTTPException(status_code=401, detail="Invalid authorization header")



import time
from typing import Iterable, List, Literal, Optional, Union
from openai.types.beta.threads import (
    Message,
    MessageDelta,
    MessageContentPartParam,
    AnnotationDelta,
    TextDeltaBlock,
    TextDelta,
)

from langchain_openai_api_bridge.assistant.adapter.openai_message_content_adapter import (
    to_openai_message_content_list,
)


def create_message(
    id: str,
    thread_id: str,
    role: Literal["user", "assistant"],
    content: Union[str, Iterable[MessageContentPartParam]] = "",
    status: Literal["in_progress", "incomplete", "completed"] = "completed",
    run_id: Optional[str] = None,
    metadata: Optional[object] = {},
) -> Message:

    return Message(
        id=id,
        thread_id=thread_id,
        role=role,
        status=status,
        object="thread.message",
        created_at=int(time.time()),
        content=to_openai_message_content_list(content=content),
        run_id=run_id,
        metadata=metadata,
    )


def create_text_message_delta(
    content: str,
    role: str,
    annotations: Optional[List[AnnotationDelta]] = [],
) -> MessageDelta:
    return MessageDelta(
        content=[
            TextDeltaBlock(
                index=0,
                type="text",
                text=TextDelta(value=content, annotations=annotations),
            )
        ],
        role=role,
    )



from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from .message import OpenAIChatMessage


class OpenAIChatCompletionRequest(BaseModel):
    model: str
    messages: List[OpenAIChatMessage]
    max_tokens: Optional[int] = 512
    temperature: Optional[float] = 0.1
    stream: Optional[bool] = False


class OpenAIChatCompletionUsage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    def default():
        return OpenAIChatCompletionUsage(
            prompt_tokens=-1, completion_tokens=-1, total_tokens=-1
        )


class OpenAIChatCompletionChoice(BaseModel):
    index: int
    message: OpenAIChatMessage
    finish_reason: Optional[str] = None


class OpenAIChatCompletionObject(BaseModel):
    id: str
    object: str = ("chat.completion",)
    created: int
    model: str
    system_fingerprint: str
    choices: List[OpenAIChatCompletionChoice]
    usage: Optional[OpenAIChatCompletionUsage]


class OpenAIChatCompletionChunkChoice(BaseModel):
    index: int
    delta: Union[OpenAIChatMessage, Dict[str, None]] = {}
    finish_reason: Optional[str]


class OpenAIChatCompletionChunkObject(BaseModel):
    id: str
    object: str
    created: int
    model: Optional[str]
    system_fingerprint: Optional[str]
    choices: List[OpenAIChatCompletionChunkChoice]
    usage: Optional[OpenAIChatCompletionUsage] = OpenAIChatCompletionUsage.default()



from typing import AsyncIterator
import uuid
from langchain_core.runnables.schema import StreamEvent

from langchain_openai_api_bridge.chat_completion.chat_completion_chunk_choice_adapter import (
    to_openai_chat_completion_chunk_object,
)
from langchain_openai_api_bridge.chat_completion.chat_completion_chunk_object_factory import (
    create_final_chat_completion_chunk_object,
)
from langchain_openai_api_bridge.core.types.openai import (
    OpenAIChatCompletionChunkObject,
)


class LangchainStreamAdapter:
    def __init__(self, llm_model: str, system_fingerprint: str = ""):
        self.llm_model = llm_model
        self.system_fingerprint = system_fingerprint

    async def ato_chat_completion_chunk_stream(
        self,
        astream_event: AsyncIterator[StreamEvent],
        id: str = str(uuid.uuid4()),
    ) -> AsyncIterator[OpenAIChatCompletionChunkObject]:
        async for event in astream_event:
            kind = event["event"]
            match kind:
                case "on_chat_model_stream":
                    chunk = to_openai_chat_completion_chunk_object(
                        event=event,
                        id=id,
                        model=self.llm_model,
                        system_fingerprint=self.system_fingerprint,
                    )
                    yield chunk

        stop_chunk = create_final_chat_completion_chunk_object(
            id=id, model=self.llm_model
        )
        yield stop_chunk



from openai.types.beta.assistant_stream_event import (
    ThreadRunCreated,
    ThreadRunCompleted,
)
from openai.types.beta.threads import (
    Run,
)

from langchain_openai_api_bridge.assistant.repository.run_repository import (
    RunRepository,
)


class ThreadRunEventHandler:

    def __init__(
        self,
        run_repository: RunRepository,
    ) -> None:
        self.run_repository = run_repository

    def on_thread_run_start(
        self, assistant_id: str, thread_id: str, model: str
    ) -> ThreadRunCreated:
        run = self.run_repository.create(
            assistant_id=assistant_id,
            thread_id=thread_id,
            model=model,
            status="in_progress",
        )

        return ThreadRunCreated(
            event="thread.run.created",
            data=run,
        )

    def on_thread_run_completed(self, run: Run) -> ThreadRunCompleted:
        completed_run = run.copy()
        completed_run.status = "completed"
        completed_run = self.run_repository.update(run=completed_run)

        return ThreadRunCompleted(
            event="thread.run.completed",
            data=completed_run,
        )



from fastapi import APIRouter, Header
from fastapi.responses import JSONResponse

from langchain_openai_api_bridge.core.agent_factory import AgentFactory
from langchain_openai_api_bridge.core.create_agent_dto import CreateAgentDto
from langchain_openai_api_bridge.chat_completion.http_stream_response_adapter import (
    HttpStreamResponseAdapter,
)
from langchain_openai_api_bridge.chat_completion.chat_completion_compatible_api import (
    ChatCompletionCompatibleAPI,
)
from langchain_openai_api_bridge.core.types.openai import OpenAIChatCompletionRequest
from langchain_openai_api_bridge.core.utils.tiny_di_container import TinyDIContainer
from langchain_openai_api_bridge.fastapi.token_getter import get_bearer_token


def create_chat_completion_router(
    tiny_di_container: TinyDIContainer,
):
    chat_completion_router = APIRouter(prefix="/chat/completions")

    @chat_completion_router.post("/")
    async def assistant_retreive_thread_messages(
        request: OpenAIChatCompletionRequest, authorization: str = Header(None)
    ):
        api_key = get_bearer_token(authorization)
        agent_factory = tiny_di_container.resolve(AgentFactory)
        create_agent_dto = CreateAgentDto(
            model=request.model,
            api_key=api_key,
            temperature=request.temperature,
        )
        llm = agent_factory.create_llm(dto=create_agent_dto)
        agent = agent_factory.create_agent(llm=llm, dto=create_agent_dto)

        adapter = ChatCompletionCompatibleAPI.from_agent(agent, create_agent_dto.model)

        response_factory = HttpStreamResponseAdapter()
        if request.stream is True:
            stream = adapter.astream(request.messages)
            return response_factory.to_streaming_response(stream)
        else:
            return JSONResponse(content=adapter.invoke(request.messages))

    return chat_completion_router


def create_openai_chat_completion_router(
    tiny_di_container: TinyDIContainer, prefix: str = ""
):
    router = create_chat_completion_router(tiny_di_container=tiny_di_container)
    open_ai_router = APIRouter(prefix=f"{prefix}/openai/v1")
    open_ai_router.include_router(router)

    return open_ai_router



from langchain_openai_api_bridge.assistant.adapter.langgraph_event_to_openai_assistant_event_stream import (
    LanggraphEventToOpenAIAssistantEventStream,
)
from langchain_openai_api_bridge.assistant.adapter.on_chat_model_end_handler import (
    OnChatModelEndHandler,
)
from langchain_openai_api_bridge.assistant.adapter.on_chat_model_stream_handler import (
    OnChatModelStreamHandler,
)
from langchain_openai_api_bridge.assistant.adapter.on_tool_end_handler import (
    OnToolEndHandler,
)
from langchain_openai_api_bridge.assistant.adapter.on_tool_start_handler import (
    OnToolStartHandler,
)
from langchain_openai_api_bridge.assistant.adapter.thread_run_event_handler import (
    ThreadRunEventHandler,
)
from langchain_openai_api_bridge.assistant.adapter.thread_to_langchain_input_messages_service import (
    ThreadToLangchainInputMessagesService,
)
from langchain_openai_api_bridge.core.utils.tiny_di_container import TinyDIContainer


def register_assistant_adapter(
    tiny_di_container: TinyDIContainer,
) -> TinyDIContainer:
    tiny_di_container.register(OnChatModelStreamHandler)
    tiny_di_container.register(OnChatModelEndHandler)
    tiny_di_container.register(ThreadRunEventHandler)
    tiny_di_container.register(OnToolStartHandler)
    tiny_di_container.register(OnToolEndHandler)

    tiny_di_container.register(ThreadToLangchainInputMessagesService)
    tiny_di_container.register(LanggraphEventToOpenAIAssistantEventStream)

    return tiny_di_container



import time
from typing import Dict, List, Optional

from langchain_openai_api_bridge.core.types.openai import (
    OpenAIChatCompletionChunkChoice,
    OpenAIChatCompletionChunkObject,
)


def create_chat_completion_chunk_object(
    id: str,
    model: str,
    system_fingerprint: Optional[str],
    choices: List[OpenAIChatCompletionChunkChoice] = [],
) -> OpenAIChatCompletionChunkObject:
    return OpenAIChatCompletionChunkObject(
        id=id,
        object="chat.completion.chunk",
        created=int(time.time()),
        model=model,
        system_fingerprint=system_fingerprint,
        choices=choices,
    )


def create_final_chat_completion_chunk_choice(
    index: int,
) -> OpenAIChatCompletionChunkChoice:
    return OpenAIChatCompletionChunkChoice(index=index, delta={}, finish_reason="stop")


def create_final_chat_completion_chunk_object(
    id: str,
    model: str = "",
    system_fingerprint: Optional[str] = None,
) -> Dict:
    return create_chat_completion_chunk_object(
        id=id,
        model=model,
        system_fingerprint=system_fingerprint,
        choices=[create_final_chat_completion_chunk_choice(index=0)],
    )



from .message import OpenAIChatMessage
from .chat_completion import (
    OpenAIChatCompletionRequest,
    OpenAIChatCompletionUsage,
    OpenAIChatCompletionChoice,
    OpenAIChatCompletionObject,
    OpenAIChatCompletionChunkChoice,
    OpenAIChatCompletionChunkObject,
)

__all__ = [
    "OpenAIChatMessage",
    "OpenAIChatCompletionRequest",
    "OpenAIChatCompletionUsage",
    "OpenAIChatCompletionChoice",
    "OpenAIChatCompletionObject",
    "OpenAIChatCompletionChunkChoice",
    "OpenAIChatCompletionChunkObject",
]



from openai.types.beta.threads import Run
import uuid
from typing import List, Optional
from langchain_openai_api_bridge.assistant.openai_run_factory import create_run
from langchain_openai_api_bridge.assistant.repository.run_repository import (
    RunRepository,
)
from openai.types.beta.threads.run import RequiredAction, RunStatus, AssistantTool


class InMemoryRunRepository(RunRepository):
    def __init__(self, data: Optional[dict[str, Run]] = None) -> None:
        self.runs = data or {}

    def create(
        self,
        assistant_id: str,
        thread_id: str,
        model: str,
        status: RunStatus,
        instructions: str = "",
        required_action: Optional[RequiredAction] = None,
        tools: List[AssistantTool] = [],
        parallel_tool_calls: bool = True,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
    ) -> Run:
        id = str(uuid.uuid4())
        run = create_run(
            id=id,
            assistant_id=assistant_id,
            thread_id=thread_id,
            model=model,
            status=status,
            instructions=instructions,
            required_action=required_action,
            tools=tools,
            parallel_tool_calls=parallel_tool_calls,
            temperature=temperature,
            top_p=top_p,
        )
        self.runs[id] = run

        return self.retreive(run_id=id)

    def retreive(self, run_id: str) -> Run:
        result = self.runs.get(run_id)
        if result is None:
            return None

        return result.copy(deep=True)

    def update(self, run: Run) -> Run:
        id = run.id
        self.runs[id] = run
        return self.retreive(run_id=id)

    def delete(self, run: Optional[Run] = None, run_id: Optional[str] = None) -> None:
        if run_id is None:
            if run is None:
                raise ValueError("At least one of run or run_id must be provided")
            id = run.id
        else:
            id = run_id

        del self.runs[id]



from typing import Optional
from pydantic import BaseModel


class CreateAgentDto(BaseModel):
    api_key: Optional[str] = None
    model: Optional[str] = None
    temperature: float = 0.7
    max_tokens: Optional[int] = None
    assistant_id: Optional[str] = ""



from .assistant_api_router import create_openai_assistant_router
from .chat_completion_router import create_openai_chat_completion_router
from .langchain_openai_api_bridge_fastapi import LangchainOpenaiApiBridgeFastAPI

__all__ = [
    "create_openai_assistant_router",
    "create_openai_chat_completion_router",
    "LangchainOpenaiApiBridgeFastAPI",
]



from langchain_openai_api_bridge.chat_completion.chat_completion_object_factory import (
    ChatCompletionObjectFactory,
)
from langchain_openai_api_bridge.chat_completion.content_adapter import (
    to_string_content,
)
from langchain_openai_api_bridge.core.role_adapter import to_openai_role
from langchain_openai_api_bridge.core.types.openai import (
    OpenAIChatCompletionChoice,
    OpenAIChatCompletionObject,
    OpenAIChatMessage,
)


class LangchainInvokeAdapter:
    def __init__(self, llm_model: str, system_fingerprint: str = ""):
        self.llm_model = llm_model
        self.system_fingerprint = system_fingerprint

    def to_chat_completion_object(self, invoke_result) -> OpenAIChatCompletionObject:
        last_message = invoke_result["messages"][-1]

        return ChatCompletionObjectFactory.create(
            id=last_message.id,
            model=self.llm_model,
            system_fingerprint=self.system_fingerprint,
            choices=[
                OpenAIChatCompletionChoice(
                    index=0,
                    message=OpenAIChatMessage(
                        role=to_openai_role(last_message.type),
                        content=to_string_content(content=last_message.content),
                    ),
                    finish_reason="stop",
                )
            ],
        )



from typing import Literal, Optional
from fastapi import APIRouter, Header

from langchain_openai_api_bridge.assistant.assistant_message_service import (
    AssistantMessageService,
)
from langchain_openai_api_bridge.assistant.assistant_run_service import (
    AssistantRunService,
)
from langchain_openai_api_bridge.assistant.assistant_stream_event_adapter import (
    AssistantStreamEventAdapter,
)
from langchain_openai_api_bridge.assistant.assistant_thread_service import (
    AssistantThreadService,
)
from langchain_openai_api_bridge.assistant.create_thread_api_dto import CreateThreadDto
from langchain_openai_api_bridge.assistant.create_thread_message_api_dto import (
    CreateThreadMessageDto,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)
from langchain_openai_api_bridge.core.agent_factory import AgentFactory
from langchain_openai_api_bridge.core.create_agent_dto import CreateAgentDto
from langchain_openai_api_bridge.core.utils.tiny_di_container import TinyDIContainer
from langchain_openai_api_bridge.fastapi.token_getter import get_bearer_token


def create_thread_router(
    tiny_di_container: TinyDIContainer,
):
    thread_router = APIRouter(prefix="/threads")

    @thread_router.post("/")
    def assistant_create_thread(create_request: CreateThreadDto):
        service = tiny_di_container.resolve(AssistantThreadService)
        return service.create(create_request)

    @thread_router.get("/")
    def assistant_list_threads(
        after: str = None,
        before: str = None,
        limit: int = 100,
        order: Literal["asc", "desc"] = None,
    ):
        service = tiny_di_container.resolve(AssistantThreadService)
        return service.list(after=after, before=before, limit=limit, order=order)

    @thread_router.get("/{thread_id}")
    def assistant_retreive_thread(thread_id: str):
        service = tiny_di_container.resolve(AssistantThreadService)
        return service.retreive(thread_id=thread_id)

    @thread_router.post("/{thread_id}")
    def assistant_update_thread(thread_id: str, request: Optional[dict] = None):
        metadata = request.get("metadata") if request else None
        service = tiny_di_container.resolve(AssistantThreadService)
        return service.update(thread_id=thread_id, metadata=metadata)

    @thread_router.delete("/{thread_id}")
    def assistant_delete_thread(thread_id: str):
        service = tiny_di_container.resolve(AssistantThreadService)
        return service.delete(thread_id=thread_id)

    @thread_router.get("/{thread_id}/messages")
    async def assistant_list_thread_messages(
        thread_id: str,
        after: str = None,
        before: str = None,
        limit: int = 100,
        order: Literal["asc", "desc"] = None,
    ):
        service = tiny_di_container.resolve(AssistantMessageService)
        messages = service.list(
            thread_id=thread_id, after=after, before=before, limit=limit, order=order
        )

        return messages

    @thread_router.get("/{thread_id}/messages/{message_id}")
    async def assistant_retreive_thread_messages(
        thread_id: str,
        message_id: str,
    ):
        service = tiny_di_container.resolve(AssistantMessageService)
        message = service.retreive(thread_id=thread_id, message_id=message_id)

        return message

    @thread_router.delete("/{thread_id}/messages/{message_id}")
    def assistant_delete_thread_messages(
        thread_id: str,
        message_id: str,
    ):
        service = tiny_di_container.resolve(AssistantMessageService)
        return service.delete(thread_id=thread_id, message_id=message_id)

    @thread_router.post("/{thread_id}/messages")
    def assistant_create_thread_messages(
        thread_id: str,
        request: CreateThreadMessageDto,
    ):
        service = tiny_di_container.resolve(AssistantMessageService)
        message = service.create(thread_id=thread_id, dto=request)

        return message

    @thread_router.post("/{thread_id}/runs")
    async def assistant_create_thread_runs(
        thread_run_dto: ThreadRunsDto,
        thread_id: str,
        authorization: str = Header(None),
        stream: bool = True,
    ):
        thread_run_dto.thread_id = thread_id

        api_key = get_bearer_token(authorization)

        agent_factory = tiny_di_container.resolve(AgentFactory)
        create_agent_dto = CreateAgentDto(
            model=thread_run_dto.model,
            api_key=api_key,
            temperature=thread_run_dto.temperature,
            assistant_id=thread_run_dto.assistant_id,
        )
        llm = agent_factory.create_llm(dto=create_agent_dto)
        agent = agent_factory.create_agent(llm=llm, dto=create_agent_dto)

        service = tiny_di_container.resolve(AssistantRunService)
        stream = service.astream(agent=agent, dto=thread_run_dto)

        response_factory = AssistantStreamEventAdapter()

        return response_factory.to_streaming_response(stream)

    return thread_router


def create_openai_assistant_router(
    tiny_di_container: TinyDIContainer, prefix: str = ""
):
    thread_router = create_thread_router(tiny_di_container=tiny_di_container)

    assistant_router = APIRouter(prefix=f"{prefix}/openai/v1")
    assistant_router.include_router(thread_router)

    return assistant_router



import json
from typing import AsyncIterator
from starlette.responses import StreamingResponse
from openai.types.beta import AssistantStreamEvent


class AssistantStreamEventAdapter:
    async def to_str_stream(
        self, assistant_event: AsyncIterator[AssistantStreamEvent]
    ) -> AsyncIterator[str]:
        async for event in assistant_event:
            str_event = self.__serialize_event(event)
            yield str_event

        done_event = self.__str_event("done", "[DONE]")
        yield done_event

    def to_streaming_response(
        self, chunk_stream: AsyncIterator[dict]
    ) -> StreamingResponse:
        str_stream = self.to_str_stream(chunk_stream)
        return StreamingResponse(
            str_stream,
            headers={
                "Content-Type": "text/event-stream",
                "Transfer-Encoding": "chunked",
            },
            media_type="application/x-ndjson",
        )

    def __serialize_event(self, event: AssistantStreamEvent):
        return self.__str_event(event.event, json.dumps(event.data.dict()))

    def __str_event(self, event: str, data: str) -> str:
        return f"event: {event}\ndata: {data}\n\n"



from langchain_core.runnables.schema import StreamEvent
from openai.types.beta import AssistantStreamEvent

from langchain_openai_api_bridge.assistant.adapter.openai_event_factory import (
    create_langchain_tool_run_step_tools_created,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)


class OnToolStartHandler:
    def __init__(self) -> None:
        pass

    def handle(
        self, event: StreamEvent, dto: ThreadRunsDto
    ) -> list[AssistantStreamEvent]:
        step_id = event["run_id"]
        name = event["name"]
        arguments = event["data"]["input"]
        metadata = event.get("metadata", None)
        tool_created_event = create_langchain_tool_run_step_tools_created(
            step_id=step_id,
            assistant_id=dto.assistant_id,
            thread_id=dto.thread_id,
            status="in_progress",
            name=name,
            arguments=arguments,
            metadata=metadata,
        )

        return [tool_created_event]



from typing import Literal, Optional
from abc import ABC, abstractmethod
from openai.types.beta import Thread, ThreadDeleted
from openai.pagination import SyncCursorPage


class ThreadRepository(ABC):

    def __init__(
        self,
    ) -> None:
        pass

    @abstractmethod
    def create(
        self,
        metadata: Optional[object] = None,
    ) -> Thread:
        # client.beta.threads.create(messages)
        pass

    @abstractmethod
    def update(
        self,
        thread_id: str,
        metadata: Optional[object] = None,
    ) -> Thread:
        # client.beta.threads.create(messages)
        pass

    @abstractmethod
    def list(
        self,
        after: str = None,
        before: str = None,
        limit: int = None,
        order: Literal["asc", "desc"] = None,
    ) -> SyncCursorPage[Thread]:
        pass

    @abstractmethod
    def retreive(self, thread_id: str) -> Thread:
        # client.beta.threads.retrieve()
        pass

    @abstractmethod
    def delete(
        self,
        thread_id: str,
    ) -> ThreadDeleted:
        # client.beta.threads.delete()
        pass



from typing import Type, TypeVar, Dict, Any, Union, Callable
import inspect

T = TypeVar("T")


class TinyDIContainer:
    def __init__(self):
        self.services: Dict[Type[Any], Any] = {}
        self.singletons: Dict[Type[Any], Any] = {}
        self.implementations: Dict[Type[Any], Type[Any]] = {}

    def register(
        self,
        cls: Type[T],
        service: Union[T, Callable[[], T], None] = None,
        *,
        singleton: bool = False,
        to: Type[T] = None
    ) -> None:
        implementation = to if to else cls
        if singleton:
            self.singletons[cls] = service if service else implementation
        else:
            self.services[cls] = service if service else implementation

        self.implementations[cls] = implementation

    def resolve(self, cls: Type[T]) -> T:
        if cls in self.singletons:
            if isinstance(self.singletons[cls], type):
                # Create and store the singleton instance
                self.singletons[cls] = self._create_instance(self.singletons[cls])
            elif callable(self.singletons[cls]):
                # Call the factory function and store the result
                self.singletons[cls] = self.singletons[cls]()
            return self.singletons[cls]

        if cls in self.services:
            if isinstance(self.services[cls], type):
                # Create a new instance every time it's resolved
                return self._create_instance(self.services[cls])
            elif callable(self.services[cls]):
                # Call the factory function
                return self.services[cls]()
            return self.services[cls]

        if cls in self.implementations:
            return self._create_instance(self.implementations[cls])

        # Automatic resolution of dependencies
        return self._create_instance(cls)

    def _create_instance(self, cls: Type[T]) -> T:
        constructor = inspect.signature(cls.__init__)
        dependencies = {
            name: self.resolve(param.annotation)
            for name, param in constructor.parameters.items()
            if name != "self" and param.annotation != param.empty
        }
        return cls(**dependencies)



from .repository import (
    MessageRepository,
    RunRepository,
    ThreadRepository,
    InMemoryMessageRepository,
    InMemoryRunRepository,
    InMemoryThreadRepository,
)


__all__ = [
    "MessageRepository",
    "RunRepository",
    "ThreadRepository",
    "InMemoryMessageRepository",
    "InMemoryRunRepository",
    "InMemoryThreadRepository",
]



import time
import uuid
from typing import Literal, Optional
from openai.types.beta import Thread, ThreadDeleted
from openai.pagination import SyncCursorPage
from .thread_repository import ThreadRepository


class InMemoryThreadRepository(ThreadRepository):
    def __init__(self, data: Optional[dict[str, Thread]] = None) -> None:
        self.threads = data or {}

    def create(self, metadata: Optional[object] = None) -> Thread:
        thread_id = str(uuid.uuid4())
        thread = self.__create_thread(thread_id=thread_id, metadata=metadata)
        self.threads[thread_id] = thread

        return self.retreive(thread_id)

    def update(
        self,
        thread_id: str,
        metadata: Optional[object] = None,
    ) -> Thread:
        if thread_id not in self.threads:
            raise ValueError(f"Thread with id {thread_id} not found")

        thread = self.threads[thread_id].copy(deep=True)
        thread.metadata = metadata
        self.threads[thread_id] = thread

        return thread

    def list(
        self,
        after: str = None,
        before: str = None,
        limit: int = None,
        order: Literal["asc", "desc"] = None,
    ) -> SyncCursorPage[Thread]:
        # Not optimal, but works well for test cases.
        # Production should use database implementation
        threads = list(self.threads.values())

        return SyncCursorPage(data=threads)

    def retreive(self, thread_id: str) -> Thread | None:
        result = self.threads.get(thread_id)
        if result is None:
            return None

        return result.copy(deep=True)

    def delete(
        self,
        thread_id: str,
    ) -> ThreadDeleted:
        thread = self.retreive(thread_id)
        if thread is None:
            return None
        del self.threads[thread_id]
        return self.__create_thread_deleted(thread_id=thread_id)

    @staticmethod
    def __create_thread(thread_id: str, metadata: Optional[object] = None) -> Thread:
        return Thread(
            id=thread_id,
            object="thread",
            created_at=int(time.time()),
            metadata=metadata,
        )

    @staticmethod
    def __create_thread_deleted(thread_id: str) -> ThreadDeleted:
        return ThreadDeleted(id=thread_id, object="thread.deleted", deleted=True)



from abc import ABC, abstractmethod
from typing import Iterable, List, Literal, Optional, Union
from openai.types.beta.threads import MessageDeleted, MessageContentPartParam
from openai.types.beta import thread_create_params
from openai.pagination import SyncCursorPage
from openai.types.beta.threads.message import Message, Attachment


class MessageRepository(ABC):

    def __init__(
        self,
    ) -> None:
        pass

    @abstractmethod
    def create(
        self,
        thread_id: str,
        role: Literal["user", "assistant"],
        content: Union[str, Iterable[MessageContentPartParam]],
        status: Literal["in_progress", "incomplete", "completed"] = "completed",
        assistant_id: Optional[str] = None,
        attachments: Optional[List[Attachment]] = None,
        run_id: Optional[str] = None,
        metadata: Optional[dict] = {},
    ) -> Message:
        pass

    @abstractmethod
    def create_many(
        self,
        thread_id: str,
        messages: List[thread_create_params.Message],
    ) -> Message:
        pass

    @abstractmethod
    def list(
        self,
        thread_id: str,
    ) -> List[Message]:
        pass

    @abstractmethod
    def listByPage(
        self,
        thread_id: str,
        after: str = None,
        before: str = None,
        limit: int = None,
        order: Literal["asc", "desc"] = None,
    ) -> SyncCursorPage[Message]:
        pass

    @abstractmethod
    def retreive(self, message_id: str, thread_id: str) -> Union[Message, None]:
        pass

    @abstractmethod
    def retreive_unique_by_run_id(
        self, run_id: str, thread_id: str
    ) -> Union[Message, None]:
        pass

    # The id is required for message delta, however, it's not necessary to hit the database
    # every time. The correlation id - run_id can be cached in this function
    @abstractmethod
    def retreive_message_id_by_run_id(
        self, run_id: str, thread_id: str
    ) -> Union[str, None]:
        pass

    @abstractmethod
    def update(self, message: Message) -> Message:
        pass

    @abstractmethod
    def delete(self, message_id: str, thread_id: str) -> MessageDeleted:
        pass



from typing import AsyncIterator
from langchain_core.runnables.schema import StreamEvent
from openai.types.beta import AssistantStreamEvent
from langchain_openai_api_bridge.assistant.adapter.on_chat_model_end_handler import (
    OnChatModelEndHandler,
)
from langchain_openai_api_bridge.assistant.adapter.on_chat_model_stream_handler import (
    OnChatModelStreamHandler,
)
from langchain_openai_api_bridge.assistant.adapter.on_tool_end_handler import (
    OnToolEndHandler,
)
from langchain_openai_api_bridge.assistant.adapter.on_tool_start_handler import (
    OnToolStartHandler,
)
from langchain_openai_api_bridge.assistant.adapter.thread_run_event_handler import (
    ThreadRunEventHandler,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)


class LanggraphEventToOpenAIAssistantEventStream:
    def __init__(
        self,
        thread_run_event_factory: ThreadRunEventHandler,
        on_chat_model_stream_handler: OnChatModelStreamHandler,
        on_chat_model_end_handler: OnChatModelEndHandler,
        on_tool_start_handler: OnToolStartHandler,
        on_tool_end_handler: OnToolEndHandler,
    ) -> None:
        self.thread_run_event_factory = thread_run_event_factory
        self.on_chat_model_stream_handler = on_chat_model_stream_handler
        self.on_chat_model_end_handler = on_chat_model_end_handler
        self.on_tool_start_handler = on_tool_start_handler
        self.on_tool_end_handler = on_tool_end_handler

    async def to_openai_assistant_event_stream(
        self,
        astream_events: AsyncIterator[StreamEvent],
        dto: ThreadRunsDto,
    ) -> AsyncIterator[AssistantStreamEvent]:

        thread_run = self.thread_run_event_factory.on_thread_run_start(
            assistant_id=dto.assistant_id,
            thread_id=dto.thread_id,
            model=dto.model,
        )

        yield thread_run

        async for event in astream_events:
            kind = event["event"]
            adapted_events: list[AssistantStreamEvent] = []
            match kind:
                case "on_chat_model_stream":
                    adapted_events += self.on_chat_model_stream_handler.handle(
                        event=event, dto=dto
                    )
                case "on_chat_model_end":
                    adapted_events += self.on_chat_model_end_handler.handle(
                        event=event, dto=dto
                    )
                case "on_tool_start":
                    adapted_events += self.on_tool_start_handler.handle(
                        event=event, dto=dto
                    )
                case "on_tool_end":
                    adapted_events += self.on_tool_end_handler.handle(
                        event=event, dto=dto
                    )
            for event in adapted_events:
                yield event

        yield self.thread_run_event_factory.on_thread_run_completed(run=thread_run.data)



import json
import time
from typing import Literal, Optional, Union
from openai.types.beta.assistant_stream_event import (
    ThreadMessageCreated,
    ThreadMessageDelta,
    MessageDeltaEvent,
    ThreadMessageCompleted,
    ThreadRunStepCreated,
    ThreadRunStepCompleted,
    ThreadRunStepDelta,
    RunStep,
)

from openai.types.beta.threads import (
    Message,
)
from openai.types.beta.threads.runs import (
    FunctionToolCall,
    ToolCallsStepDetails,
)
from openai.types.beta.threads.runs import function_tool_call

from langchain_openai_api_bridge.assistant.adapter.openai_message_factory import (
    create_text_message_delta,
)


def create_thread_message_created_event(message: Message) -> ThreadMessageCreated:
    return ThreadMessageCreated(
        event="thread.message.created",
        data=message,
    )


def create_text_thread_message_delta(
    message_id: str, content: str, role: str
) -> ThreadMessageDelta:
    return ThreadMessageDelta(
        event="thread.message.delta",
        data=MessageDeltaEvent(
            id=message_id,
            delta=create_text_message_delta(content=content, role=role),
            object="thread.message.delta",
        ),
    )


def create_thread_message_completed(
    message: Message,
) -> ThreadMessageCompleted:

    return ThreadMessageCompleted(
        event="thread.message.completed",
        data=message,
    )


def create_langchain_tool_run_step_tools_created(
    step_id: str,
    assistant_id: str,
    thread_id: str,
    status: Literal["in_progress", "cancelled", "failed", "completed", "expired"],
    metadata: Optional[object] = None,
    name: Optional[str] = None,
    arguments: Optional[dict[object]] = None,
    output: Optional[str] = None,
) -> ThreadRunStepDelta:

    return ThreadRunStepCreated(
        event="thread.run.step.created",
        data=create_langchain_tool_run_step(
            step_id=step_id,
            assistant_id=assistant_id,
            thread_id=thread_id,
            status=status,
            metadata=metadata,
            name=name,
            arguments=arguments,
            output=output,
        ),
    )


def create_langchain_tool_thread_run_step_completed(
    step_id: str,
    assistant_id: str,
    thread_id: str,
    status: Literal["in_progress", "cancelled", "failed", "completed", "expired"],
    metadata: Optional[object] = None,
    name: Optional[str] = None,
    arguments: Optional[Union[dict[object], float, str]] = None,
    output: Optional[Union[dict[object], float, str]] = None,
) -> ThreadRunStepCompleted:
    return ThreadRunStepCompleted(
        event="thread.run.step.completed",
        data=create_langchain_tool_run_step(
            step_id=step_id,
            assistant_id=assistant_id,
            thread_id=thread_id,
            status=status,
            metadata=metadata,
            name=name,
            arguments=arguments,
            output=output,
        ),
    )


def create_langchain_tool_run_step(
    step_id: str,
    assistant_id: str,
    thread_id: str,
    status: Literal["in_progress", "cancelled", "failed", "completed", "expired"],
    metadata: Optional[object] = None,
    name: Optional[str] = None,
    arguments: Optional[Union[dict[object], float, str]] = None,
    output: Optional[Union[dict[object], float, str]] = None,
) -> RunStep:
    return RunStep(
        id=step_id,
        assistant_id=assistant_id,
        created_at=int(time.time()),
        metadata=metadata,
        object="thread.run.step",
        run_id=step_id,
        status=status,
        step_details=ToolCallsStepDetails(
            type="tool_calls",
            tool_calls=[
                create_langchain_tool_tool_call(
                    id=step_id,
                    name=name,
                    arguments=arguments,
                    output=output,
                )
            ],
        ),
        thread_id=thread_id,
        type="tool_calls",
    )


def create_langchain_tool_tool_call(
    id: str,
    name: Optional[str] = "",
    arguments: Optional[Union[dict[object], float, str]] = None,
    output: Optional[Union[dict[object], float, str]] = None,
) -> FunctionToolCall:

    function = create_langchain_function(name=name, arguments=arguments, output=output)
    return function_tool_call.FunctionToolCall(
        id=id,
        function=function,
        type="function",
    )


def create_langchain_function(
    name: Optional[str] = "",
    arguments: Optional[Union[dict[object], float, str]] = None,
    output: Optional[Union[dict[object], float, str]] = None,
) -> function_tool_call.Function:
    arguments_json = json.dumps(arguments) if arguments else None
    output_json = json.dumps(output) if output else None
    return function_tool_call.Function(
        name=name, arguments=arguments_json, output=output_json
    )



from typing import Iterable, List, Literal, Optional, Union
import uuid

from langchain_openai_api_bridge.assistant.adapter.openai_message_factory import create_message
from .message_repository import (
    MessageRepository,
)
from openai.types.beta import thread_create_params
from openai.types.beta.threads import (
    Message,
    MessageDeleted,
    MessageContentPartParam,
)
from openai.pagination import SyncCursorPage


class InMemoryMessageRepository(MessageRepository):
    def __init__(self, data: Optional[dict[str, Message]] = None) -> None:
        self.messages = data or {}

    def create(
        self,
        thread_id: str,
        role: Literal["user", "assistant"],
        content: Union[str, Iterable[MessageContentPartParam]],
        status: Literal["in_progress", "incomplete", "completed"] = "completed",
        run_id: Optional[str] = None,
        metadata: Optional[object] = {},
    ) -> Message:
        id = str(uuid.uuid4())
        message = create_message(
            id=id,
            thread_id=thread_id,
            role=role,
            content=content,
            status=status,
            run_id=run_id,
            metadata=metadata,
        )
        self.messages[id] = message

        return self.retreive(message_id=id, thread_id=thread_id)

    def create_many(
        self,
        thread_id: str,
        messages: List[thread_create_params.Message],
    ) -> List[Message]:
        created_messages = []
        for message in messages:
            created_message = self.create(
                thread_id=thread_id,
                role=message["role"],
                content=message["content"],
                status=message.get("status", "completed"),
            )
            created_messages.append(created_message)
        return created_messages

    def list(
        self,
        thread_id: str,
    ) -> List[Message]:
        # Not optimal, but works well for test cases.
        # Production should use database implementation
        messages = [
            message.copy(deep=True)
            for message in self.messages.values()
            if message.thread_id == thread_id
        ]

        return messages

    def listByPage(
        self,
        thread_id: str,
        after: str = None,
        before: str = None,
        limit: int = None,
        order: Literal["asc", "desc"] = None,
    ) -> SyncCursorPage[Message]:
        messages = self.list(thread_id=thread_id)

        return SyncCursorPage(data=messages)

    # thread_id is not needed for in-memory implementation
    def retreive(self, message_id: str, thread_id: str) -> Union[Message, None]:
        result = self.messages.get(message_id)
        if result is None:
            return None

        return result.copy(deep=True)

    def retreive_unique_by_run_id(self, run_id: str, thread_id: str) -> Message:
        messages = [
            message.copy(deep=True)
            for message in self.messages.values()
            if message.thread_id == thread_id and message.run_id == run_id
        ]

        if not messages:
            return None

        return messages[0]

    def retreive_message_id_by_run_id(
        self, run_id: str, thread_id: str
    ) -> Union[str, None]:
        message = self.retreive_unique_by_run_id(run_id=run_id, thread_id=thread_id)

        if message is None:
            return None

        return message.id

    def update(self, message: Message) -> Message:
        id = message.id
        self.messages[id] = message
        return self.retreive(message_id=id, thread_id=message.thread_id)

    def delete(self, message_id: str, thread_id: str) -> MessageDeleted:
        message = self.retreive(thread_id=thread_id, message_id=message_id)
        if message is None:
            return None
        del self.messages[message_id]
        return self.__create_message_deleted(message_id=message_id)

    @staticmethod
    def __create_message_deleted(message_id: str) -> MessageDeleted:
        return MessageDeleted(
            id=message_id, object="thread.message.deleted", deleted=True
        )



from typing import Callable, Union
from fastapi import FastAPI

from langchain_openai_api_bridge.assistant.adapter.container import (
    register_assistant_adapter,
)
from langchain_openai_api_bridge.assistant.assistant_message_service import (
    AssistantMessageService,
)
from langchain_openai_api_bridge.assistant.assistant_run_service import (
    AssistantRunService,
)
from langchain_openai_api_bridge.assistant.assistant_thread_service import (
    AssistantThreadService,
)
from langchain_openai_api_bridge.assistant.repository.message_repository import (
    MessageRepository,
)
from langchain_openai_api_bridge.assistant.repository.run_repository import (
    RunRepository,
)
from langchain_openai_api_bridge.assistant.repository.thread_repository import (
    ThreadRepository,
)
from langchain_openai_api_bridge.core.agent_factory import AgentFactory
from langchain_openai_api_bridge.core.langchain_openai_api_bridge import (
    LangchainOpenaiApiBridge,
)
from langchain_openai_api_bridge.fastapi.assistant_api_router import (
    create_openai_assistant_router,
)
from langchain_openai_api_bridge.fastapi.chat_completion_router import (
    create_openai_chat_completion_router,
)


class LangchainOpenaiApiBridgeFastAPI(LangchainOpenaiApiBridge):
    def __init__(
        self,
        app: FastAPI,
        agent_factory_provider: Union[Callable[[], AgentFactory], AgentFactory],
    ) -> None:
        super().__init__(agent_factory_provider=agent_factory_provider)
        self.app = app

    def bind_openai_assistant_api(
        self,
        thread_repository_provider: Union[Callable[[], ThreadRepository]],
        message_repository_provider: Union[Callable[[], MessageRepository]],
        run_repository_provider: Union[Callable[[], RunRepository]],
        prefix: str = "",
    ) -> None:

        self.tiny_di_container.register(AssistantThreadService)
        self.tiny_di_container.register(AssistantMessageService)
        self.tiny_di_container.register(AssistantRunService)
        self.tiny_di_container.register(ThreadRepository, to=thread_repository_provider)
        self.tiny_di_container.register(
            MessageRepository, to=message_repository_provider
        )
        self.tiny_di_container.register(RunRepository, to=run_repository_provider)
        register_assistant_adapter(self.tiny_di_container)

        assistant_router = create_openai_assistant_router(
            tiny_di_container=self.tiny_di_container,
            prefix=prefix,
        )

        self.app.include_router(assistant_router)

    def bind_openai_chat_completion(self, prefix: str = "") -> None:
        chat_completion_router = create_openai_chat_completion_router(
            self.tiny_di_container, prefix=prefix
        )

        self.app.include_router(chat_completion_router)



import time
from typing import List, Optional
from openai.types.beta.threads import Run
from openai.types.beta.threads.run import RequiredAction, RunStatus, AssistantTool


def create_run(
    id: str,
    assistant_id: str,
    thread_id: str,
    model: str,
    status: RunStatus,
    instructions: str = "",
    required_action: Optional[RequiredAction] = None,
    parallel_tool_calls: bool = True,
    tools: List[AssistantTool] = [],
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
) -> Run:
    return Run(
        id=id,
        assistant_id=assistant_id,
        thread_id=thread_id,
        model=model,
        created_at=int(time.time()),
        instructions=instructions,
        object="thread.run",
        parallel_tool_calls=parallel_tool_calls,
        required_action=required_action,
        status=status,
        tools=tools,
        temperature=temperature,
        top_p=top_p,
    )



from typing import Optional
from langchain_core.runnables.schema import StreamEvent

from langchain_openai_api_bridge.chat_completion.chat_completion_chunk_object_factory import (
    create_chat_completion_chunk_object,
)
from langchain_openai_api_bridge.chat_completion.content_adapter import (
    to_string_content,
)
from langchain_openai_api_bridge.core.types.openai import (
    OpenAIChatCompletionChunkChoice,
    OpenAIChatCompletionChunkObject,
    OpenAIChatMessage,
)


def to_openai_chat_message(
    event: StreamEvent,
    role: str = "assistant",
) -> OpenAIChatMessage:
    content = event["data"]["chunk"].content
    return OpenAIChatMessage(content=to_string_content(content), role=role)


def to_openai_chat_completion_chunk_choice(
    event: StreamEvent,
    index: int = 0,
    role: str = "assistant",
    finish_reason: Optional[str] = None,
) -> OpenAIChatCompletionChunkChoice:
    message = to_openai_chat_message(event, role)

    return OpenAIChatCompletionChunkChoice(
        index=index,
        delta=message,
        finish_reason=finish_reason,
    )


def to_openai_chat_completion_chunk_object(
    event: StreamEvent,
    id: str = "",
    model: str = "",
    system_fingerprint: Optional[str] = None,
    role: str = "assistant",
    finish_reason: Optional[str] = None,
) -> OpenAIChatCompletionChunkObject:

    choice1 = to_openai_chat_completion_chunk_choice(
        event, index=0, role=role, finish_reason=finish_reason
    )

    return create_chat_completion_chunk_object(
        id=id,
        model=model,
        system_fingerprint=system_fingerprint,
        choices=[choice1],
    )



from typing import Iterable, List, Type, Union
from openai import BaseModel
from openai.types.beta.threads import (
    MessageContentPartParam,
    TextContentBlock,
    ImageFile,
    ImageURL,
    ImageFileContentBlock,
    ImageURLContentBlock,
    Text,
)
from openai.types.beta.threads.message import MessageContent


def to_openai_message_content(
    content: Union[str, MessageContentPartParam] = ""
) -> MessageContent:

    if isinstance(content, str):
        return TextContentBlock(text=Text(value=content, annotations=[]), type="text")
    elif isinstance(content, dict):
        if content["type"] == "text":
            return TextContentBlock(
                type="text",
                text=Text(value=content["text"], annotations=[]),
            )
        if content["type"] == "image_file":
            return ImageFileContentBlock(
                type="image_file",
                image_file=ImageFile(
                    file_id=content["image_file"]["file_id"],
                    detail=content["image_file"]["detail"],
                ),
            )
        if content["type"] == "image_url":
            return ImageURLContentBlock(
                type="image_url",
                image_url=ImageURL(
                    url=content["image_url"]["url"],
                    detail=content["image_url"]["detail"],
                ),
            )
    elif (
        isinstance(content, TextContentBlock)
        or isinstance(content, ImageFileContentBlock)
        or isinstance(content, ImageURLContentBlock)
    ):
        return content


def to_openai_message_content_list(
    content: Union[str, Iterable[MessageContentPartParam]] = ""
) -> List[MessageContent]:
    if isinstance(content, str):
        return [to_openai_message_content(content)]
    elif isinstance(content, Iterable):
        return [to_openai_message_content(item) for item in content]


def deserialize_message_content(data: dict) -> MessageContent:
    type_to_model: dict[str, Type[BaseModel]] = {
        "image_file": ImageFileContentBlock,
        "image_url": ImageURLContentBlock,
        "text": TextContentBlock,
    }

    content_type = data["type"]
    model_cls = type_to_model[content_type]
    return model_cls.parse_obj(data)



from abc import ABC, abstractmethod
from langgraph.graph.graph import CompiledGraph
from langchain_core.language_models import BaseChatModel
from langchain_openai_api_bridge.core.create_agent_dto import CreateAgentDto


class AgentFactory(ABC):

    @abstractmethod
    def create_agent(self, llm: BaseChatModel, dto: CreateAgentDto) -> CompiledGraph:
        pass

    @abstractmethod
    def create_llm(self, dto: CreateAgentDto) -> CompiledGraph:
        pass



from langchain_core.runnables.schema import StreamEvent
from openai.types.beta import AssistantStreamEvent

from langchain_openai_api_bridge.assistant.adapter.openai_event_factory import (
    create_langchain_tool_thread_run_step_completed,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)


class OnToolEndHandler:
    def __init__(self) -> None:
        pass

    def handle(
        self, event: StreamEvent, dto: ThreadRunsDto
    ) -> list[AssistantStreamEvent]:
        step_id = event["run_id"]
        name = event["name"]
        arguments = event["data"].get("input", None)
        output = event["data"].get("output", None)
        metadata = event["metadata"]
        tool_completed_event = create_langchain_tool_thread_run_step_completed(
            step_id=step_id,
            assistant_id=dto.assistant_id,
            thread_id=dto.thread_id,
            status="in_progress",
            name=name,
            arguments=arguments,
            output=output,
            metadata=metadata,
        )

        return [tool_completed_event]



from typing import Literal, Optional
from openai.types.beta import Thread, ThreadDeleted
from langchain_openai_api_bridge.assistant.create_thread_api_dto import CreateThreadDto
from langchain_openai_api_bridge.assistant.repository.message_repository import (
    MessageRepository,
)
from langchain_openai_api_bridge.assistant.repository.thread_repository import (
    ThreadRepository,
)
from openai.pagination import SyncCursorPage


class AssistantThreadService:

    def __init__(
        self,
        thread_repository: ThreadRepository,
        message_repository: MessageRepository,
    ) -> None:
        self.thread_repository = thread_repository
        self.message_repository = message_repository

    def create(
        self,
        dto: CreateThreadDto,
    ) -> Thread:
        # Reference:
        # client.beta.threads.create(messages)

        thread = self.thread_repository.create(metadata=dto.metadata)

        if len(dto.messages) > 0:
            self.message_repository.create_many(
                thread_id=thread.id, messages=dto.messages
            )

        return thread

    def list(
        self,
        after: str = None,
        before: str = None,
        limit: int = None,
        order: Literal["asc", "desc"] = None,
    ) -> SyncCursorPage[Thread]:
        return self.thread_repository.list(
            after=after, before=before, limit=limit, order=order
        )

    def retreive(self, thread_id: str) -> Thread:
        # Reference:
        # client.beta.threads.retrieve()

        return self.thread_repository.retreive(thread_id=thread_id)

    def update(self, thread_id: str, metadata: Optional[object] = None) -> Thread:
        # Reference:
        # client.beta.threads.update()

        return self.thread_repository.update(thread_id=thread_id, metadata=metadata)

    def delete(
        self,
        thread_id: str,
    ) -> ThreadDeleted:
        # Reference:
        # client.beta.threads.delete()
        return self.thread_repository.delete(thread_id=thread_id)



from langchain_core.runnables.schema import StreamEvent
from openai.types.beta import AssistantStreamEvent
from langchain_openai_api_bridge.assistant.adapter.openai_event_factory import (
    create_text_thread_message_delta,
    create_thread_message_created_event,
)
from langchain_openai_api_bridge.assistant.create_thread_runs_api_dto import (
    ThreadRunsDto,
)
from langchain_openai_api_bridge.assistant.repository.message_repository import (
    MessageRepository,
)


class OnChatModelStreamHandler:
    def __init__(
        self,
        thread_message_repository: MessageRepository,
    ):
        self.thread_message_repository = thread_message_repository

    def handle(self, event: StreamEvent, dto: ThreadRunsDto) -> AssistantStreamEvent:
        events = []
        chunk = event["data"]["chunk"]
        content: str = chunk.content
        run_id = event["run_id"]

        if content is None or content == "":
            return events

        message_id = self.thread_message_repository.retreive_message_id_by_run_id(
            run_id=run_id, thread_id=dto.thread_id
        )
        if message_id is None:
            created_message = self.thread_message_repository.create(
                thread_id=dto.thread_id,
                role="assistant",
                content="",
                status="in_progress",
                run_id=run_id,
            )
            message_id = created_message.id
            events.append(create_thread_message_created_event(message=created_message))

        events.append(
            create_text_thread_message_delta(
                message_id=message_id, content=content, role="assistant"
            )
        )

        return events
